use std::io::Read;
use bytes::Buf;
use thiserror::Error;

#[derive(Error, Debug)]
pub enum HeaderParseError {
    #[error("Not enough data")]
    NotEnoughData,
    #[error("Invalid magic header")]
    InvalidMagicHeader,
    #[error("Invalid page size")]
    InvalidPageSize,
    #[error("Invalid file format version")]
    InvalidFileFormatVersion,
    #[error("Invalid text encoding")]
    InvalidTextEncoding,
    #[error("Invalid maximum embedded payload fraction")]
    InvalidMaximumEmbeddedPayloadFraction,
    #[error("Invalid minimum embedded payload fraction")]
    InvalidMinimumEmbeddedPayloadFraction,
    #[error("Invalid leaf payload fraction")]
    InvalidLeafPayloadFraction,
}

const MAGIC_HEADER: &str = "SQLite format 3\0";

#[derive(Debug, Clone, Copy)]
pub enum SqliteFileFormatVersion {
    Legacy = 1,
    WAL = 2,
}

#[derive(Debug, Clone, Copy)]
enum SqliteTextEncoding {
    Utf8 = 1,
    Utf16le = 2,
    Utf16be = 3,
}

const MAX_EMBEDDED_PAYLOAD_FRACTION: u8 = 64;
const MIN_EMBEDDED_PAYLOAD_FRACTION: u8 = 32;
const LEAF_PAYLOAD_FRACTION: u8 = 32;

#[derive(Debug, Clone, Copy)]
pub struct Version {
    major: u8,
    minor: u8,
    patch: u8,
}

impl std::fmt::Display for Version {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}.{}.{}", self.major, self.minor, self.patch)
    }
}

impl Version {
    pub fn from_sqlite_version_number(mut version_number: u32) -> Self {
        let major = (version_number / 1000000) as u8;
        version_number %= 1000000;
        let minor = (version_number / 1000) as u8;
        version_number %= 1000;
        let patch = version_number as u8;
        Version {
            major,
            minor,
            patch,
        }
    }
}

#[derive(Debug)]
pub struct SqliteHeader {
    /**
     * The database page size in bytes. Must be a power of two between 512 and 65536 inclusive.
     */
    pub page_size: u32,
    /*
     * The file format write version. 1 for legacy, 2 for WAL.
     */
    pub file_format_write_version: SqliteFileFormatVersion,
    pub file_format_read_version: SqliteFileFormatVersion,
    /**
     * The reserved space at the end of each page. Usually 0.
     */
    pub reserved_space: u8,
    /**
     * file change counter.
     */
    pub file_change_counter: u32,
    /**
     * The size of the database in pages. The "in-header database size" is the number of pages in the database according to the most recent commit.
     * This might be different from the "on-disk database size" if the database has been growing and the size of the database file has not yet been updated.
     */
    pub database_size: u32,
    /**
     * The page number of the first freelist trunk page.
     */
    pub first_freelist_trunk_page: u32,
    /**
     * The total number of freelist pages.
     */
    pub total_freelist_pages: u32,
    /**
     * The schema cookie.
     */
    pub schema_cookie: u32,
    /**
     * The schema format number.
     * New database files created by SQLite use format 4 by default.
     * If the database is completely empty, then the schema format number is 0.
     */
    pub schema_format_number: u32,
    /**
     * The default page cache size in pages. The value is a suggestion only.
     */
    pub default_page_cache_size: u32,

    pub largest_root_b_tree_page_number: u32,
    /**
     * The database text encoding. 1 for UTF-8, 2 for UTF-16le, 3 for UTF-16be.
     */
    pub text_encoding: SqliteTextEncoding,
    /**
     * The user version.
     * sqlite3 does not use this value internally. It is available for use by applications.
     */
    pub user_version: u32,

    pub incremental_vacuum_mode: u32,
    /**
     * The application ID.
     * sqlite3 does not use this value internally. It is available for use by applications.
     */
    pub application_id: u32,
    /**
     * The version-valid-for number.
     */
    pub version_valid_for: u32,
    /**
     * The SQLite version number.
     */
    pub sqlite_version_number: Version,
}

impl TryFrom<u8> for SqliteFileFormatVersion {
    type Error = HeaderParseError;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        match value {
            1 => Ok(SqliteFileFormatVersion::Legacy),
            2 => Ok(SqliteFileFormatVersion::WAL),
            _ => Err(HeaderParseError::InvalidFileFormatVersion),
        }
    }
}

impl TryFrom<u32> for SqliteTextEncoding {
    type Error = HeaderParseError;

    fn try_from(value: u32) -> Result<Self, Self::Error> {
        match value {
            1 => Ok(SqliteTextEncoding::Utf8),
            2 => Ok(SqliteTextEncoding::Utf16le),
            3 => Ok(SqliteTextEncoding::Utf16be),
            _ => Err(HeaderParseError::InvalidTextEncoding),
        }
    }
}

impl SqliteHeader {
    fn parse_page_size(page_size: u16) -> Result<u32, HeaderParseError> {
        if page_size == 1 {
            Ok(65536)
        } else {
            if page_size < 512 || page_size > 32768 {
                Err(HeaderParseError::InvalidPageSize)
            }
            else {
                Ok(page_size as u32)
            }
        }
    }
    fn parse_maximum_embedded_payload_fraction(fraction: u8) -> Result<u8, HeaderParseError> {
        if (fraction < MIN_EMBEDDED_PAYLOAD_FRACTION) || (fraction > MAX_EMBEDDED_PAYLOAD_FRACTION) {
            Err(HeaderParseError::InvalidMaximumEmbeddedPayloadFraction)
        }
        else {
            Ok(fraction)
        }
    }

    fn parse_minimum_embedded_payload_fraction(fraction: u8) -> Result<u8, HeaderParseError> {
        if (fraction < MIN_EMBEDDED_PAYLOAD_FRACTION) || (fraction > MAX_EMBEDDED_PAYLOAD_FRACTION) {
            Err(HeaderParseError::InvalidMinimumEmbeddedPayloadFraction)
        }
        else {
            Ok(fraction)
        }
    }

    fn parse_leaf_payload_fraction(fraction: u8) -> Result<u8, HeaderParseError> {
        if (fraction < LEAF_PAYLOAD_FRACTION) || (fraction > MAX_EMBEDDED_PAYLOAD_FRACTION) {
            Err(HeaderParseError::InvalidLeafPayloadFraction)
        }
        else {
            Ok(fraction)
        }
    }

    pub fn read_from_reader<R: Read>(reader: &mut R) -> Result<Self, HeaderParseError> {
        let mut buffer = [0u8; 100];
        reader.read_exact(&mut buffer).map_err(|_| HeaderParseError::NotEnoughData)?;
        Self::read(&buffer)
    }

    pub fn read(buffer: &[u8]) -> Result<Self, HeaderParseError> {
        let mut reader = buffer;
        // check file size
        if buffer.len() < 100 {
            return Err(HeaderParseError::NotEnoughData);
        }
        // check magic header
        let magic_header = &reader.copy_to_bytes(MAGIC_HEADER.len());
        let magic_header = std::str::from_utf8(&magic_header).map_err(
            |_| HeaderParseError::InvalidMagicHeader
        )?;

        if magic_header != MAGIC_HEADER {
            return Err(HeaderParseError::InvalidMagicHeader);
        }

        let page_size =  SqliteHeader::parse_page_size(reader.get_u16())?;
        let file_format_write_version = SqliteFileFormatVersion::try_from(reader.get_u8())?;
        let file_format_read_version = SqliteFileFormatVersion::try_from(reader.get_u8())?;
        let reserved_space = reader.get_u8();
        let maximum_embedded_payload_fraction = reader.get_u8();
        SqliteHeader::parse_maximum_embedded_payload_fraction(maximum_embedded_payload_fraction)?;
        let minimum_embedded_payload_fraction = reader.get_u8();
        SqliteHeader::parse_minimum_embedded_payload_fraction(minimum_embedded_payload_fraction)?;
        let leaf_payload_fraction = reader.get_u8();
        SqliteHeader::parse_leaf_payload_fraction(leaf_payload_fraction)?;

        let file_change_counter = reader.get_u32();
        let database_size = reader.get_u32();
        let first_freelist_trunk_page = reader.get_u32();
        let total_freelist_pages = reader.get_u32();
        let schema_cookie = reader.get_u32();
        let schema_format_number = reader.get_u32();
        let default_page_cache_size = reader.get_u32();
        let largest_root_b_tree_page_number = reader.get_u32();
        let text_encoding = SqliteTextEncoding::try_from(reader.get_u32())?;
        let user_version = reader.get_u32();
        let incremental_vacuum_mode = reader.get_u32();
        let application_id = reader.get_u32();
        // reserved space
        reader.advance(20);
        let version_valid_for = reader.get_u32();
        let sqlite_version_number = reader.get_u32();
        let sqlite_version_number = Version::from_sqlite_version_number(sqlite_version_number);

        Ok(Self {
            page_size,
            file_format_write_version,
            file_format_read_version,
            reserved_space,
            file_change_counter,
            database_size,
            first_freelist_trunk_page,
            total_freelist_pages,
            schema_cookie,
            schema_format_number,
            default_page_cache_size,
            largest_root_b_tree_page_number,
            text_encoding,
            user_version,
            incremental_vacuum_mode,
            application_id,
            version_valid_for,
            sqlite_version_number,
        })
    }
}

#[cfg(test)]
mod test {
    #[test]
    fn test_version_from_sqlite_version_number() {
        use super::Version;
        assert_eq!(Version::from_sqlite_version_number(3008000).to_string(), "3.8.0");
        assert_eq!(Version::from_sqlite_version_number(3010000).to_string(), "3.10.0");
        assert_eq!(Version::from_sqlite_version_number(3011000).to_string(), "3.11.0");
    }

    #[test]
    fn test_parse_sqlite_page_size() {
        use super::SqliteHeader;
        assert_eq!(SqliteHeader::parse_page_size(1).unwrap(), 65536);
        assert_eq!(SqliteHeader::parse_page_size(512).unwrap(), 512);
        assert_eq!(SqliteHeader::parse_page_size(32768).unwrap(), 32768);
        assert!(SqliteHeader::parse_page_size(511).is_err());
        assert!(SqliteHeader::parse_page_size(32769).is_err());
    }

    #[test]
    fn test_parse_maximum_embedded_payload_fraction() {
        use super::SqliteHeader;
        assert_eq!(SqliteHeader::parse_maximum_embedded_payload_fraction(32).unwrap(), 32);
        assert_eq!(SqliteHeader::parse_maximum_embedded_payload_fraction(64).unwrap(), 64);
        assert!(SqliteHeader::parse_maximum_embedded_payload_fraction(31).is_err());
        assert!(SqliteHeader::parse_maximum_embedded_payload_fraction(65).is_err());
    }

    #[test]
    fn test_parse_minimum_embedded_payload_fraction() {
        use super::SqliteHeader;
        assert_eq!(SqliteHeader::parse_minimum_embedded_payload_fraction(32).unwrap(), 32);
        assert_eq!(SqliteHeader::parse_minimum_embedded_payload_fraction(64).unwrap(), 64);
        assert!(SqliteHeader::parse_minimum_embedded_payload_fraction(31).is_err());
        assert!(SqliteHeader::parse_minimum_embedded_payload_fraction(65).is_err());
    }
}