use std::io::{Read, Seek};

use bytes::{buf, Buf};
use thiserror::Error;

struct FreeTrunkPage {
    /**
     * The next trunk page in the list of free pages.
     * If this is the last trunk page in the list, this field is 0.
     */
    next_trunk_page: u32,
    /**
     * vector of free list leaf pages in this trunk page.
     */
    free_list: Vec<u32>,
// A bug in SQLite versions prior to 3.6.0 (2008-07-16) caused the database to be reported as corrupt 
// if any of the last 6 entries in the freelist trunk page array contained non-zero values. 
// Newer versions of SQLite do not have this problem. However, newer versions of SQLite still avoid 
// using the last six entries in the freelist trunk page array in order that database files created 
// by newer versions of SQLite can be read by older versions of SQLite.
}

impl FreeTrunkPage {
    pub fn read_from_reader<R: std::io::Read>(reader: &mut R, page_size: u32) -> Result<Self, Error> {
        let mut buf = vec![0; page_size as usize];
        reader.read_exact(&mut buf)?;
        let mut buf = buf.as_slice();
        let next_trunk_page = buf.get_u32();
        let count = buf.get_u32();
        let mut free_list = Vec::new();
        for _ in 0..count {
            free_list.push(buf.get_u32());
        }
        Ok(FreeTrunkPage {
            next_trunk_page,
            free_list,
        })
    }
    pub fn get_next_trunk_page(&self) -> u32 {
        self.next_trunk_page
    }
    pub fn get_free_list(&self) -> &Vec<u32> {
        &self.free_list
    }
}

/**
 * A free list leaf page.
 * Should contain no information.
 */
struct FreeLeapPage;

/**
 * B-tree page.
 */
#[derive(Debug)]
pub enum BTreePage {
    /**
     * Internal node.
     */
    InternalNode(InternalNode),
    /**
     * Leaf node.
     */
    LeafNode(LeafNode),
}

#[derive(Debug)]
enum NodeType {
    Table,
    Index,
}

#[derive(Debug)]
pub struct InternalNode {
    node_type: NodeType,

    /**
     * start of the cells on the page
     */
    first_free_cell: u16,
    /**
     * number of cells on the page
     */
    cell_count: u16,
    /**
     * cell content area start
     */
    data_start: u16,
    /**
     * the number of fragmented free bytes within the cell content area. 
     */
    fragment_count: u8,
    last_child_cell: u32,

    child_cells: Vec<u32>,
}

#[derive(Debug)]
pub struct LeafNode {
    node_type: NodeType,
    /**
     * start of the cells on the page
     */
    first_free_cell: u16,
    /**
     * number of cells on the page
     */
    cell_count: u16,
    /**
     * cell content area start
     */
    data_start: u16,
    /**
     * the number of fragmented free bytes within the cell content area. 
     */
    fragment_count: u8,

    pub data_cells: Vec<DataCell>,
}

#[derive(Debug)]
pub struct DataCell {
    pub row_id: u64,
    pub data: Vec<u8>,
}

#[derive(Debug)]
pub enum Value {
    Null,
    Integer(i64),
    Real(f64),
    Text(String),
    Blob(Vec<u8>),
}

impl Value {
    pub fn from_bytes(data: &mut &[u8], serial_type: u64) -> Value {
        let buf = data;
        match serial_type {
            0 => Value::Null,
            1 => Value::Integer(buf.get_u8() as i64),
            2 => Value::Integer(buf.get_u16() as i64),
            3 => Value::Integer({
                let mut value = 0;
                for i in 0..3 {
                    value |= (buf.get_u8() as i64) << (i * 8);
                }
                value
            } as i64),
            4 => Value::Integer(buf.get_u32() as i64),
            5 => Value::Integer({
                let mut value = 0;
                for i in 0..6 {
                    value |= (buf.get_u8() as i64) << (i * 8);
                }
                value
            } as i64),
            6 => Value::Integer(buf.get_u64() as i64),
            7 => Value::Real(buf.get_f64()),
            8 => Value::Integer(0),
            9 => Value::Integer(1),
            // Reserved for internal use.  
            10|11 => panic!("Invalid serial type: {}", serial_type),
            st if st >= 12 && st % 2 == 0 => {
                let len = (st - 12) / 2;
                let blob = buf[0..len as usize].to_vec();
                buf.advance(len as usize);
                Value::Blob(blob)
            }
            st if st >= 13 && st % 2 == 1 => {
                let len = (st - 13) / 2;
                let text = buf[0..len as usize].to_vec();
                buf.advance(len as usize);
                // println!("text: {:?} {:?}", text, len);
                Value::Text(String::from_utf8(text).unwrap())
            },
            _ => panic!("Invalid serial type: {}", serial_type),
        }
    }
}

#[derive(Debug, Error)]
pub enum Error {
    #[error("IO error: {0}")]
    IoError(#[from] std::io::Error),
    #[error("Invalid page type: {0}")]
    InvalidPageType(u8),
    #[error("Invalid page size: {0}")]
    InvalidPageSize(u32),
    #[error("Invalid cell count: {0}")]
    InvalidCellCount(u16),
    #[error("Invalid right child page: {0}")]
    InvalidRightChildPage(u32),
    #[error("Invalid child page: {0}")]
    InvalidChildPage(u32),
    #[error("Invalid cell pointer: {0}")]
    InvalidCellPointer(u16),
}

impl BTreePage {
    pub fn read_from_reader<R: Read + Seek> (
        reader: &mut R,
        page_size: u32,
    ) -> Result<BTreePage, Error> {
        let mut buf = vec![0; page_size as usize];
        reader.read_exact(&mut buf)?;
        Self::read(&buf, false)
    }
    pub fn read(
        buffer: &[u8],
        is_first_page: bool,
    ) -> Result<BTreePage, Error> {
        let mut buf = buffer;
        if is_first_page {
            // skip the header
            buf = &buffer[100..];
        }
        let page_type = buf.get_u8();
        let first_free_cell = buf.get_u16();
        let cell_count = buf.get_u16();
        let data_start = buf.get_u16();
        let fragment_count = buf.get_u8();

        match page_type {
            2 | 5 => {
                let last_child_cell = if page_type == 2 {
                    buf.get_u32()
                } else {
                    0 
                };
                
                let mut cell_pointers = Vec::new();
                for _ in 0..cell_count {
                    cell_pointers.push(buf.get_u16());
                }

                let mut child = Vec::new();
                for offset in cell_pointers {
                    let offset = offset as u32;
                    let offset = offset;
                    buf = &buffer[offset as usize..];
                    let child_page = buf.get_u32();
                    child.push(child_page);
                }

                Ok(BTreePage::InternalNode(InternalNode {
                    cell_count,
                    first_free_cell,
                    data_start,
                    fragment_count,
                    last_child_cell,
                    child_cells: child,
                    node_type: if page_type == 2 { NodeType::Table } else { NodeType::Index },
                }))
            }
            10 | 13 => {
                // buf.get_u32();
                let mut cell_pointers = Vec::new();
                for _ in 0..cell_count {
                    cell_pointers.push(buf.get_u16());
                }
                // println!("cell_pointers: {:?}", cell_pointers);
                let mut data_cells = Vec::new();
                for offset in cell_pointers {
                    println!("offset: {:#02x}", offset);
                    let offset = offset as u32;
                    let offset = offset;
                    buf = &buffer[(offset) as usize..];
                    
                    let (data_len, _) = varint_read_from_buf(&mut buf);
                    let (row_id, _) = varint_read_from_buf(&mut buf);
                    println!("data_len: {}, row_id: {}", data_len, row_id);
                    let data = buf[..data_len as usize].to_vec();
                    data_cells.push(DataCell {
                        row_id,
                        data,
                    });
                }


                Ok(BTreePage::LeafNode(LeafNode {
                    node_type: if page_type == 10 { NodeType::Table } else { NodeType::Index },
                    cell_count,
                    first_free_cell,
                    data_start,
                    fragment_count,
                    data_cells,
                }))
            }
            _ => Err(Error::InvalidPageType(page_type)),
        }
    }
    
}

pub fn varint_read_from_buf(buf: &mut &[u8]) -> (u64, u32) {
    let mut byte = buf.get_u8();
    let mut value: u64 = (byte & 0x7F) as u64;
    let mut i = 1;
    while byte & 0x80 != 0 && i < 9 {
        byte = buf.get_u8();
        value <<= 7;
        value |= (byte & 0x7F) as u64;
        i += 1;
    }
    if i == 8 && byte & 0x80 != 0 {
        byte = buf.get_u8();
        value <<= 8;
        value |= byte as u64;
        i += 1;
    };
    (value, i)
}

#[cfg(test)]
mod test {
    #[test]
    fn test_varint_read_from_buf() {
        let buf = [0x82, 0x00];
        let mut buf = buf.as_ref();
        let (value, i) = super::varint_read_from_buf(&mut buf);
        assert_eq!(value, 0x100);
        assert_eq!(i, 2);
    }

    #[test]
    fn test_varint_read_from_buf_2() {
        let buf = [0x82, 0x01];
        let mut buf = buf.as_ref();
        let (value, i) = super::varint_read_from_buf(&mut buf);
        assert_eq!(value, 0x101);
        assert_eq!(i, 2);
    }

    #[test]
    fn test_varint_read_from_buf_3() {
        let buf = [0x81, 0x91, 0xd1, 0xac, 0x78];
        let mut buf = buf.as_ref();
        let (value, i) = super::varint_read_from_buf(&mut buf);
        assert_eq!(value, 0x12345678);
        assert_eq!(i, 5);
    }
}